from django.contrib import admin

from huscy.attributes import models, services


@admin.register(models.AttributeSchema)
class AttributeSchemaAdmin(admin.ModelAdmin):
    list_display = 'id', 'schema'

    def save_model(self, request, attribute_schema, form, change):
        services.update_attribute_schema(attribute_schema.schema)

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(models.AttributeSet)
class AttributeSetAdmin(admin.ModelAdmin):
    list_display = 'pseudonym', '_attribute_schema', 'attributes'

    def _attribute_schema(self, attribute_set):
        return attribute_set.attribute_schema.id

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False



from django.shortcuts import get_object_or_404
from rest_framework.generics import RetrieveUpdateAPIView, UpdateAPIView
from rest_framework.exceptions import PermissionDenied
from rest_framework.permissions import DjangoModelPermissions, IsAuthenticated
from rest_framework.response import Response

from huscy.attributes import serializer, services
from huscy.attributes.models import AttributeSchema, AttributeSet
from huscy.attributes.permissions import (
    MigrateAttributeSetPermission,
    ReadOnly,
    RetrieveAttributeSetPermission,
    UpdateAttributeSetPermission,
)
from huscy.subjects.models import Subject


class AttributeSchemaView(RetrieveUpdateAPIView):
    permission_classes = (DjangoModelPermissions | ReadOnly, )
    queryset = AttributeSchema.objects.all()
    serializer_class = serializer.AttributeSchemaSerializer

    def get_object(self):
        return services.get_attribute_schema()


class BaseAttributeSetView:

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        self.subject = get_object_or_404(Subject, pk=self.kwargs['subject_pk'])

    def get_object(self):
        return services.get_attribute_set(self.subject)


class AttributeSetView(BaseAttributeSetView, RetrieveUpdateAPIView):
    permission_classes = (
        IsAuthenticated,
        RetrieveAttributeSetPermission | UpdateAttributeSetPermission
    )
    serializer_class = serializer.AttributeSetSerializer

    def update(self, request, *args, **kwargs):
        attribute_set = self.get_object()

        serializer = self.get_serializer(attribute_set, data=request.data)
        serializer.is_valid(raise_exception=True)

        for node, node_description in attribute_set.attribute_schema.schema['properties'].items():
            if (node in serializer.validated_data['attributes'] and
                    node_description['type'] == 'object' and
                    not self.request.user.has_perm(f'attributes.change_attribute_category_{node}')):
                raise PermissionDenied('You don\'t have the permission to update attribute '
                                       f'category {node}')

        self.perform_update(serializer)

        return Response(serializer.data)

    def retrieve(self, request, *args, **kwargs):
        attribute_set = self.get_object()

        for node, node_description in attribute_set.attribute_schema.schema['properties'].items():
            if all([node_description['type'] == 'object',
                    node in attribute_set.attributes,
                    not request.user.has_perm(f'attributes.view_attribute_category_{node}')]):
                attribute_set.attributes.pop(node)

        serializer = self.get_serializer(attribute_set)
        return Response(serializer.data)


class MigrateAttributeSetView(BaseAttributeSetView, UpdateAPIView):
    permission_classes = IsAuthenticated, MigrateAttributeSetPermission
    queryset = AttributeSet.objects.none()
    serializer_class = serializer.MigrateAttributeSetSerializer



from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from huscy.attributes import models, services


class AttributeSchemaSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.AttributeSchema
        fields = (
            'id',
            'created_at',
            'schema',
        )

    def create(self, validated_data):
        return services.update_attribute_schema(**validated_data)


class AttributeSetSerializer(serializers.ModelSerializer):
    attribute_schema = AttributeSchemaSerializer(read_only=True)

    class Meta:
        model = models.AttributeSet
        fields = (
            'attribute_schema',
            'attributes',
        )

    def update(self, attribute_set, validated_data):
        return services.update_attribute_set(attribute_set, **validated_data)


class MigrateAttributeSetSerializer(serializers.ModelSerializer):
    attribute_schema = serializers.PrimaryKeyRelatedField(
        queryset=models.AttributeSchema.objects.all(),
        required=False,
    )
    attributes = serializers.JSONField(required=False)

    class Meta:
        model = models.AttributeSet
        fields = (
            'attribute_schema',
            'attributes',
        )

    def update(self, attribute_set, validated_data):
        try:
            return services.migrate_attribute_set(attribute_set, **validated_data)
        except services.AttributeSetMigrationError as e:
            raise ValidationError(str(e))



from operator import attrgetter

import jsonschema
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.db.transaction import atomic

from huscy.attributes.models import AttributeSchema, AttributeSet
from huscy.pseudonyms.services import get_or_create_pseudonym


class AttributeSetMigrationError(Exception):
    pass


def get_attribute_schema(version=None):
    if version is None:
        return AttributeSchema.objects.latest('pk')
    else:
        return AttributeSchema.objects.get(pk=version)


def get_attribute_set(subject):
    pseudonym = get_or_create_pseudonym(subject, 'attributes.attributeset')
    return _get_or_create_attribute_set(pseudonym.code)


def _get_or_create_attribute_set(pseudonym):
    attribute_set, _created = AttributeSet.objects.select_related('attribute_schema').get_or_create(
        pseudonym=pseudonym,
        defaults=dict(
            attribute_schema=AttributeSchema.objects.latest('id'),
            attributes={},
        ),
    )
    return attribute_set


def migrate_attribute_set(attribute_set, attribute_schema=None, attributes=None):
    if attribute_schema is None:
        attribute_schema = AttributeSchema.objects.latest('id')

    if attribute_schema.pk <= attribute_set.attribute_schema.pk:
        raise AttributeSetMigrationError('New version for attribute schema must be greater than '
                                         'current attribute schema version.')

    attribute_set.attribute_schema = attribute_schema

    if attributes is not None:
        attribute_set.attributes = attributes

    jsonschema.validate(attribute_set.attributes, attribute_schema.schema)

    attribute_set.save()
    return attribute_set


@atomic
def update_attribute_schema(schema):
    attribute_schema = AttributeSchema.objects.create(schema=schema)
    created_permissions = _create_attribute_category_permissions(schema)
    _delete_orphaned_attribute_category_permissions(created_permissions)
    return attribute_schema


def _create_attribute_category_permissions(schema):
    content_type = ContentType.objects.get_for_model(AttributeSchema)

    permissions = []
    for name, value in schema['properties'].items():
        if value['type'] == 'object':
            permissions.append(_create_attribute_category_read_permission(content_type, name))
            permissions.append(_create_attribute_category_write_permission(content_type, name))
    return permissions


def _create_attribute_category_read_permission(content_type, attribute_category_name):
    permission, created = Permission.objects.get_or_create(
        codename=f'view_attribute_category_{attribute_category_name}',
        name=f'Can view attribute category {attribute_category_name}',
        content_type=content_type
    )
    return permission


def _create_attribute_category_write_permission(content_type, attribute_category_name):
    permission, created = Permission.objects.get_or_create(
        codename=f'change_attribute_category_{attribute_category_name}',
        name=f'Can change attribute category {attribute_category_name}',
        content_type=content_type
    )
    return permission


def _delete_orphaned_attribute_category_permissions(created_permissions):
    queryset = Permission.objects
    queryset = queryset.filter(Q(codename__startswith='change_attribute_category_') |
                               Q(codename__startswith='view_attribute_category_'))
    queryset = queryset.exclude(codename__in=map(attrgetter('codename'), created_permissions))
    queryset.delete()


def update_attribute_set(attribute_set, attributes):
    _dict_merge(attribute_set.attributes, attributes)
    jsonschema.validate(attribute_set.attributes, attribute_set.attribute_schema.schema)
    attribute_set.save(update_fields=['attributes'])
    return attribute_set


def _dict_merge(source_dict, merge_dict):
    for key, value in merge_dict.items():
        if key in source_dict and isinstance(source_dict[key], dict) and isinstance(value, dict):
            _dict_merge(source_dict[key], value)
        else:
            source_dict[key] = value



from django.db import models
from django.utils.translation import gettext_lazy as _


class AttributeSchema(models.Model):
    schema = models.JSONField(_('Schema'))
    created_at = models.DateTimeField(_('Created at'), auto_now_add=True, editable=False)

    class Meta:
        verbose_name = _('Attribute schema')
        verbose_name_plural = _('Attribute schemas')


class AttributeSet(models.Model):
    pseudonym = models.CharField(_('Pseudonym'), max_length=128, primary_key=True, editable=False)
    attribute_schema = models.ForeignKey(AttributeSchema, on_delete=models.PROTECT,
                                         verbose_name=_('Attribute schema'))
    attributes = models.JSONField(_('Attributes'))

    class Meta:
        permissions = [
            ('migrate_attributeset', _('Can migrate attribute set')),
        ]
        verbose_name = _('Attribute set')
        verbose_name_plural = _('Attribute sets')



from django.db import migrations, models
import django.db.models.deletion


def create_initial_attribute_schema(apps, schema_editor):
    AttributeSchema = apps.get_model("attributes", "AttributeSchema")

    AttributeSchema.objects.create(
        schema=dict(type='object', properties=dict())
    )


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AttributeSchema',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('schema', models.JSONField(verbose_name='Schema')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
            ],
            options={
                'verbose_name': 'Attribute schema',
                'verbose_name_plural': 'Attribute schemas',
            },
        ),
        migrations.CreateModel(
            name='AttributeSet',
            fields=[
                ('pseudonym', models.CharField(editable=False, max_length=128, primary_key=True, serialize=False, verbose_name='Pseudonym')),
                ('attributes', models.JSONField(verbose_name='Attributes')),
                ('attribute_schema', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='attributes.attributeschema', verbose_name='Attribute schema')),
            ],
            options={
                'verbose_name': 'Attribute set',
                'verbose_name_plural': 'Attribute sets',
                'permissions': [('migrate_attributeset', 'Can migrate attribute set')],
            },
        ),

        migrations.RunPython(create_initial_attribute_schema)
    ]



from rest_framework.permissions import BasePermission, SAFE_METHODS


class MigrateAttributeSetPermission(BasePermission):
    def has_permission(self, request, view):
        return all([
            request.method == 'PUT',
            request.user.has_perm('attributes.migrate_attributeset'),
        ])


class RetrieveAttributeSetPermission(BasePermission):
    def has_permission(self, request, view):
        return all([
            request.method == 'GET',
            request.user.has_perm('attributes.view_attributeset'),
        ])


class UpdateAttributeSetPermission(BasePermission):
    def has_permission(self, request, view):
        return all([
            request.method == 'PUT',
            request.user.has_perms(['subjects.change_subject', 'attributes.change_attributeset'])
        ])


class ReadOnly(BasePermission):
    def has_permission(self, request, view):
        return request.user.is_authenticated and request.method in SAFE_METHODS



from django.urls import path

from . import views


urlpatterns = [
    path(
        'api/attributeschema/',
        views.AttributeSchemaView.as_view(),
        name='attributeschema',
    ),
    path(
        'api/subjects/<uuid:subject_pk>/attributeset/migrate/',
        views.MigrateAttributeSetView.as_view(),
        name='migrate-attributeset',
    ),
    path(
        'api/subjects/<uuid:subject_pk>/attributeset/',
        views.AttributeSetView.as_view(),
        name='attributeset',
    ),
]
